import concurrent.futures
import os
from rdkit import Chem
from rdkit.Chem import AllChem, rdMolDescriptors, rdForceFieldHelpers, rdDistGeom
from itertools import repeat
from datetime import datetime
from nrgrank import process_ligands
import subprocess
from pathlib import Path
import csv
import argparse


def get_delimiter(file_path, bytes_to_read=4096):
    sniffer = csv.Sniffer()
    data = open(file_path, "r").read(bytes_to_read)
    delimiter = sniffer.sniff(data).delimiter
    return delimiter


def read_column_from_csv(file_path, column_number, delimiter, has_header=True):
    column_values = []
    with open(file_path, mode='r', encoding='utf-8') as file:
        reader = csv.reader(file, delimiter=delimiter)
        if has_header:
            next(reader, None)
        for row in reader:
            if len(row) > column_number:
                column_values.append(row[column_number])

    return column_values


def generate_conformers(molecule_smile, molecule_name, no_conformers, mol_weight_max=None, heavy_atoms_min=None):
    etkdg = rdDistGeom.ETKDGv3()
    # === optional settings ===
    # etkdg.maxAttempts = 10
    # etkdg.pruneRmsThresh = 0.5
    # etkdg.numThreads = 10
    # https://greglandrum.github.io/rdkit-blog/posts/2023-03-02-clustering-conformers.html
    etkdg.randomSeed = 0xa700f
    etkdg.verbose = False
    etkdg.useRandomCoords = True
    molecule = Chem.MolFromSmiles(molecule_smile)
    try:
        frags = Chem.GetMolFrags(molecule, asMols=True, sanitizeFrags=False)
    except:
        print('Error getting fragment for: ', molecule_name)
        frags = molecule
        if frags is None:
            return None
    molecule = max(frags, key=lambda frag: frag.GetNumAtoms())
    if mol_weight_max:
        mol_weight = rdMolDescriptors.CalcExactMolWt(molecule)
        if mol_weight > mol_weight_max:
            return None
    if heavy_atoms_min:
        num_heavy_atoms = molecule.GetNumHeavyAtoms()
        if num_heavy_atoms <= heavy_atoms_min:
            return None

    mol = Chem.AddHs(molecule, addCoords=True)
    if no_conformers == 1:
        try:
            AllChem.EmbedMolecule(mol, params=etkdg)
        except Exception as e:
            print('=====================================')
            print(f'Error: {e}\n Molecule: {molecule_name}\n')
            print('=====================================')
            return None
    else:
        AllChem.EmbedMultipleConfs(mol, no_conformers, params=etkdg)
    mol.SetProp("_Name", molecule_name)

    return mol


def read_args():
    parser = argparse.ArgumentParser(description="Process and convert chemical data.")

    parser.add_argument(
        "-s",
        "--smiles_path",
        type=str,
        help="Path to the SMILES file."
    )
    parser.add_argument(
        "-sc",
        "--smiles_column_number",
        type=int,
        required=True,
        help="Number of the column containing smiles (Starts at 0). Example: -1 for last column.",
    )
    parser.add_argument(
        "-nc",
        "--name_column_number",
        type=int,
        required=True,
        help="Number of the column containing names (Starts at 0). Example: -1 for last column.",
    )
    parser.add_argument(
        "-o",
        "--output_folder_path",
        type=str,
        default=None,
        help="Path to the custom output folder. Use 'None' if not specified.",
    )
    parser.add_argument(
        "--optimize",
        action="store_true",
        help="Enable optimization. Defaults to False.",
    )
    parser.add_argument(
        "-dc",
        "--no_convert",
        action="store_false",
        dest="convert",
        help="Disable conversion from SDF to MOL2. Defaults to True.",
    )
    parser.add_argument(
        "-p",
        "--no_preprocess",
        action="store_false",
        dest="preprocess",
        help="Disable preprocessing. Defaults to True.",
    )
    parser.add_argument(
        "-mw",
        "--molecular_weight_max",
        type=int,
        default=0,
        help="Maximum molecular weight. Defaults to 0 which will not filter for MW",
    )
    parser.add_argument(
        "-ha",
        "--heavy_atoms_min",
        type=int,
        default=0,
        help="Minimum number of heavy atoms. Defaults to 0 which will not filter for this setting",
    )

    args = parser.parse_args()

    main(
        smiles_path=args.smiles_path,
        smiles_column_number=args.smiles_column_number,
        name_column_number=args.name_column_number,
        output_folder_path=args.output_folder_path,
        optimize=args.optimize,
        convert=args.convert,
        preprocess=args.preprocess,
        molecular_weight_max=args.molecular_weight_max,
        heavy_atoms_min=args.heavy_atoms_min
    )


def main(smiles_path, smiles_column_number, name_column_number, output_folder_path, conformers_per_molecule=1, optimize=False, convert=True,
         preprocess=False, molecular_weight_max=None, heavy_atoms_min=None):
    root_software_path = Path(__file__).resolve().parents[1]
    os.chdir(root_software_path)

    if conformers_per_molecule <= 0:
        exit("Number of conformers must be greater than 0.")

    if not output_folder_path:
        output_folder_path = os.path.join(os.path.dirname(smiles_path), f"{os.path.basename(smiles_path).split('.')[0]}_conformers")
    if not os.path.isdir(output_folder_path):
        os.mkdir(output_folder_path)

    sdf_output_file = os.path.join(output_folder_path, f"{os.path.splitext(os.path.basename(smiles_path))[0]}")
    if conformers_per_molecule > 1:
        sdf_output_file += f"_{conformers_per_molecule}_conf"
    if optimize:
        sdf_output_file += "_optimized"
    sdf_output_file += '.sdf'
    mol2_output_file = os.path.splitext(sdf_output_file)[0] + '.mol2'

    writer = AllChem.SDWriter(sdf_output_file)

    delimiter = get_delimiter(smiles_path, bytes_to_read=4096)
    molecule_smiles_list = read_column_from_csv(smiles_path, smiles_column_number, delimiter, has_header=True)
    molecule_name_list = read_column_from_csv(smiles_path, name_column_number, delimiter, has_header=True)

    with concurrent.futures.ThreadPoolExecutor() as executor:
        for mol in executor.map(generate_conformers, molecule_smiles_list, molecule_name_list,
                                repeat(conformers_per_molecule), repeat(molecular_weight_max), repeat(heavy_atoms_min)):
            if mol is not None:
                for cid in range(mol.GetNumConformers()):
                    if optimize:
                        Chem.rdForceFieldHelpers.MMFFOptimizeMoleculeConfs(mol, cid)
                    mol = Chem.RemoveHs(mol)
                    writer.write(mol, cid)
    AllChem.SDWriter.close(writer)
    print("Finished generating conformers @ ", datetime.now())

    if convert:
        print("converting to mol2")
        open_babel_command = f"obabel \"{sdf_output_file}\" -O \"{mol2_output_file}\" ---errorlevel 1"
        print(f'obabel command: {open_babel_command}')
        subprocess.run(open_babel_command, shell=True, check=True)
        os.remove(sdf_output_file)

    if preprocess:
        process_ligands(ligand_path=mol2_output_file)


if __name__ == '__main__':
    read_args()



import os
import numpy as np
from nrgrank.general_functions import get_radius_number, load_rad_dict
import argparse
import pickle
import re


def load_atoms_mol2(filename, save_path, ligand_type='ligand'):
    coord_start = 0
    max_atoms = 0
    n_atoms = 0
    n_molecules = 0
    n_unique_molecules = 0
    same_molec_counter = 1
    molecule_name_list = []
    rad_dict = load_rad_dict()
    with open(filename) as f:
        lines = f.readlines()

    for counter, line in enumerate(lines):
        if line.startswith('@<TRIPOS>MOLECULE'):
            n_molecules += 1
            molecule_name = lines[counter+1][0:-1]
            if molecule_name != "\n":
                molec_suffix = "_0"
                if n_molecules > 1 and molecule_name_list[n_molecules-2].split("_")[0] == molecule_name:
                    molec_suffix = f"_{same_molec_counter}"
                    same_molec_counter += 1
                else:
                    same_molec_counter = 1
                molecule_name_list.append(lines[counter+1][0:-1] + molec_suffix)
                if molec_suffix == "_0":
                    n_unique_molecules += 1
            else:
                exit("Error when reading molecule name")
        elif line.startswith("@<TRIPOS>ATOM"):
            coord_start = 1
            if max_atoms < n_atoms:
                max_atoms = n_atoms
            n_atoms = 0
        if coord_start == 1:
            if line.startswith("@<TRIPOS>ATOM") is False:
                if line[0] == "@":
                    coord_start = 0
                elif line.split()[1][0] != "H":
                    n_atoms += 1
    if max_atoms < n_atoms:
        max_atoms = n_atoms

    n_atom_array = np.zeros(n_molecules, dtype=np.int32)
    atoms_xyz = np.full((n_molecules, max_atoms, 3), 9999, dtype=np.float32)
    atoms_type = np.full((n_molecules, max_atoms), -1, dtype=np.int32)

    molecule_counter = -1
    atom_counter = 0
    coord_start = 0
    atom_name_list = []
    temp_atom_name_list = []
    atoms_name_count = {}
    for counter, line in enumerate(lines):
        if line.startswith('@<TRIPOS>MOLECULE'):
            atoms_name_count = {}
            molecule_counter += 1
            if molecule_counter > 0:
                n_atom_array[molecule_counter-1] = atom_counter
                atom_counter = 0
                coord_start = 0
                atom_name_list.append(temp_atom_name_list)
                temp_atom_name_list = []
        if line.startswith('@<TRIPOS>ATOM') and line[0] != "\n":
            coord_start = 1
        if coord_start == 1 and line.startswith('@<TRIPOS>ATOM') is False:
            if line[0] != '@':
                line = line.split()
                if line[5][0] != 'H':
                    atoms_xyz[molecule_counter][atom_counter] = np.array([float(line[2]),
                                                                          float(line[3]),
                                                                          float(line[4])])
                    atom_type = line[5]
                    atoms_type_temp, _ = get_radius_number(atom_type, rad_dict)
                    atoms_type[molecule_counter][atom_counter] = atoms_type_temp
                    atm_name = atom_type.split(".")[0]
                    if atm_name in atoms_name_count:
                        atoms_name_count[atm_name] += 1
                    else:
                        atoms_name_count[atm_name] = 1
                    temp_atom_name_list.append(f"{atm_name}{atoms_name_count[atm_name]}")
                    atom_counter += 1
            else:
                coord_start = 0
    n_atom_array[-1] = atom_counter
    atom_name_list.append(temp_atom_name_list)
    np.save(os.path.join(save_path, f"{ligand_type}_atom_xyz"), atoms_xyz)
    np.save(os.path.join(save_path, f"{ligand_type}_atom_type"), atoms_type)
    with open(os.path.join(save_path, f"{ligand_type}_molecule_name.pkl"), 'wb') as f:
        pickle.dump(molecule_name_list, f)
    with open(os.path.join(save_path, f"{ligand_type}_atom_name.pkl"), 'wb') as f:
        pickle.dump(atom_name_list, f)
    np.save(os.path.join(save_path, f"{ligand_type}_atoms_num_per_ligand"), n_atom_array)
    if ligand_type != 'ligand':
        np.save(os.path.join(save_path, f"{ligand_type}_ligand_count"), np.array([n_unique_molecules]))


def get_suffix(conf_num):
    suffix = ""
    if conf_num != 0:
        suffix = f"_{conf_num}_conf"
    return suffix


def get_suffix_search_in_file_name(filepath):
    pattern = r'(_\d+_conf)'
    match = re.search(pattern, filepath)
    return match.group(1)


def get_args():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-l", '--ligand_path', type=str, help='Path to folder containing ligand folders')
    parser.add_argument("-t", '--ligand_type', type=str, help='Ligand type (not required)')
    parser.add_argument("-c", '--conformers_per_molecule', type=int, help='Number of conformers per molecule')
    parser.add_argument("-o", '--output_dir', type=str, help='Output directory')

    args = parser.parse_args()
    ligand_file_path = args.ligand_path
    ligand_type = args.ligand_type
    conformers_per_molecule = args.conformers_per_molecule
    output_dir = args.output_dir
    main(ligand_path=ligand_file_path, ligand_type=ligand_type, conformers_per_molecule=conformers_per_molecule, output_dir=output_dir)


def main(ligand_path, conformers_per_molecule, overwrite=False, ligand_type='ligand', output_dir=None):
    if os.path.isfile(ligand_path):
        if ligand_path.find('_conf') != -1:
            suffix = get_suffix_search_in_file_name(ligand_path)
        else:
            suffix = get_suffix(conformers_per_molecule)
        if output_dir is None:
            output_dir = os.path.dirname(ligand_path)
        output_folder = os.path.join(output_dir, f"preprocessed_ligands{suffix}")
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        elif os.listdir(output_folder) and not overwrite:
            print(f'{output_folder} is not empty... Skipping. \nUse overwrite=True to overwrite.')
        else:
            load_atoms_mol2(ligand_path, output_folder, ligand_type=ligand_type)
        return output_folder
    else:
        exit(f'Argument used for ligand_path is not a file: {ligand_path}')


if __name__ == "__main__":
    get_args()



import os
import shutil
import numpy as np
from numba import njit
import timeit
import argparse
import math as m
import pickle
from nrgrank.general_functions import write_pdb

# def njit(njit):
#     return njit


@njit
def center_coords(ligand_atoms_xyz, list_size):
    length = list_size
    centered_coord = np.zeros((list_size, 3), dtype=np.float32)
    sum_x = np.sum(ligand_atoms_xyz[:, 0]) / length
    sum_y = np.sum(ligand_atoms_xyz[:, 1]) / length
    sum_z = np.sum(ligand_atoms_xyz[:, 2]) / length
    centroid_coords = np.array([sum_x, sum_y, sum_z])
    for i in range(len(ligand_atoms_xyz)):
        centered_coord[i] = ligand_atoms_xyz[i] - centroid_coords
    return centered_coord


def load_ligands(target_path, ligand_type, start, end, conf_num, path_to_ligands=None):
    if not path_to_ligands:
        if conf_num == 0:
            print('Ligands are in an old path and conf number is 0')
            path_to_ligands = os.path.join(target_path, f"preprocessed_ligands")
            if not os.path.isdir(path_to_ligands):
                exit('Could not find ligands folder')
        else:
            ligand_folder = f"preprocessed_ligands_{conf_num}_conf"
            path_to_ligands = os.path.join(target_path, ligand_folder)

    with open(os.path.join(path_to_ligands, f"{ligand_type}_atom_name.pkl"), 'rb') as f:
        atom_name = pickle.load(f)[start:end].copy()
    with open(os.path.join(path_to_ligands, f"{ligand_type}_molecule_name.pkl"), 'rb') as f:
        molecule_name = pickle.load(f)[start:end].copy()
    atom_type = np.load(os.path.join(path_to_ligands, f"{ligand_type}_atom_type.npy"), mmap_mode='r')[start:end].copy()
    atom_xyz = np.load(os.path.join(path_to_ligands, f"{ligand_type}_atom_xyz.npy"), mmap_mode='r')[start:end].copy()
    atoms_num_per_ligand = np.load(os.path.join(path_to_ligands, f"{ligand_type}_atoms_num_per_ligand.npy"),
                                   mmap_mode='r')[start:end].copy()
    ligand_count = len(atom_type)
    return atom_name, atom_type, atom_xyz, molecule_name, atoms_num_per_ligand, ligand_count


def Rx(theta):
    return np.matrix([[1, 0, 0], [0, m.cos(theta), -m.sin(theta)], [0, m.sin(theta), m.cos(theta)]])


def Ry(theta):
    return np.matrix([[m.cos(theta), 0, m.sin(theta)], [0, 1, 0], [-m.sin(theta), 0, m.cos(theta)]])


def Rz(theta):
    return np.matrix([[m.cos(theta), -m.sin(theta), 0], [m.sin(theta), m.cos(theta), 0], [0, 0, 1]])


def rotate_ligand(ligand_atoms_xyz, n_rotations):
    centered_ligand_atoms_xyz = center_coords(ligand_atoms_xyz, len(ligand_atoms_xyz))
    single_rotation = 360/n_rotations
    rotation_counter = 0
    rotated_ligand_coord_list = np.zeros((n_rotations**3, len(centered_ligand_atoms_xyz), 3), dtype=np.float32)
    for x_rot_count in range(n_rotations):
        for y_rot_count in range(n_rotations):
            for z_rot_count in range(n_rotations):
                x = np.deg2rad(x_rot_count*single_rotation)
                y = np.deg2rad(y_rot_count*single_rotation)
                z = np.deg2rad(z_rot_count*single_rotation)
                rotation_matrix = Rx(x) * Ry(y) * Rz(z)
                for i, coord in enumerate(centered_ligand_atoms_xyz):
                    rotated_ligand_coord_list[rotation_counter][i] = np.array(np.dot(rotation_matrix, coord))
                rotation_counter += 1
    rotated_ligand_coord_list_unique = np.unique(rotated_ligand_coord_list, axis=0)
    return rotated_ligand_coord_list_unique


@njit
def get_cf(lig_pose, point, cf_size_list, precalc_cf_list, ligand_atoms_types, default_cf, cell_width, min_xyz):
    cf = 0.0
    lig_pose = lig_pose + point
    x_index_array = ((lig_pose[:, 0] - min_xyz[0]) / cell_width).astype(np.int32)
    y_index_array = ((lig_pose[:, 1] - min_xyz[1]) / cell_width).astype(np.int32)
    z_index_array = ((lig_pose[:, 2] - min_xyz[2]) / cell_width).astype(np.int32)
    if np.min(x_index_array) < 0 or np.min(y_index_array) < 0 or np.min(z_index_array) < 0:
        cf = default_cf
    elif np.max(x_index_array) >= cf_size_list[0] or np.max(y_index_array) >= cf_size_list[1] or np.max(z_index_array) >= cf_size_list[2]:
        cf = default_cf
    else:
        for counter, _ in enumerate(x_index_array):
            temp_cf = precalc_cf_list[x_index_array[counter]][y_index_array[counter]][z_index_array[counter]][ligand_atoms_types[counter]-1]
            if temp_cf == default_cf:
                cf = default_cf
                break
            else:
                cf += temp_cf
    return cf


@njit
def get_cf_with_clash(lig_pose, point, load_range_list, grid_spacing, cf_size_list, load_cf_list, ligand_atoms_types,
                      default_cf, cell_width, min_xyz, clash_list, clash_list_size, num_atoms):
    ###### CHECK CLASH ######
    x_index_array = np.empty_like(lig_pose[:, 0])
    np.round(((lig_pose[:, 0] + point[0] - load_range_list[0][0]) / grid_spacing), 0, x_index_array)  # .astype(np.int32)
    y_index_array = np.empty_like(lig_pose[:, 1])
    np.round(((lig_pose[:, 1] + point[1] - load_range_list[1][0]) / grid_spacing), 0, y_index_array)  # .astype(np.int32)
    z_index_array = np.empty_like(lig_pose[:, 2])
    np.round(((lig_pose[:, 2] + point[2] - load_range_list[2][0]) / grid_spacing), 0, z_index_array)  # .astype(np.int32)
    x_index_array = x_index_array.astype(np.int32)
    y_index_array = y_index_array.astype(np.int32)
    z_index_array = z_index_array.astype(np.int32)
    x_index_array[x_index_array == cf_size_list[0]] -= 1
    y_index_array[y_index_array == cf_size_list[1]] -= 1
    z_index_array[z_index_array == cf_size_list[2]] -= 1
    if np.min(x_index_array) < 0 or np.min(y_index_array) < 0 or np.min(z_index_array) < 0:
        return default_cf
    elif np.max(x_index_array) >= clash_list_size[0] or np.max(y_index_array) >= clash_list_size[1] or np.max(z_index_array) >= clash_list_size[2]:
        return default_cf
    else:
        for number in np.arange(0, num_atoms, 1):
            clash_detect = clash_list[x_index_array[number]][y_index_array[number]][z_index_array[number]]
            if clash_detect:
                return default_cf
        else:
            cf = 0.0
            lig_pose = lig_pose + point
            x_index_array = ((lig_pose[:, 0] - min_xyz[0]) / cell_width).astype(np.int32)
            y_index_array = ((lig_pose[:, 1] - min_xyz[1]) / cell_width).astype(np.int32)
            z_index_array = ((lig_pose[:, 2] - min_xyz[2]) / cell_width).astype(np.int32)
            for counter, _ in enumerate(x_index_array):
                temp_cf = load_cf_list[x_index_array[counter]][y_index_array[counter]][z_index_array[counter]][ligand_atoms_types[counter]-1]
                if temp_cf == default_cf:
                    cf = default_cf
                    break
                else:
                    cf += temp_cf
            return cf


@njit
def get_cf_main(binding_site_grid, ligand_orientations, cf_size_list, n_cf_evals, load_cf_list, ligand_atoms_types,
                default_cf, cell_width, min_xyz):
    cfs_list = np.zeros((n_cf_evals, 3), dtype=np.float32)
    counter = 0
    for point_index, point in enumerate(binding_site_grid):
        for pose_index, lig_pose in enumerate(ligand_orientations):
            cf = get_cf(lig_pose, point, cf_size_list, load_cf_list, ligand_atoms_types, default_cf, cell_width, min_xyz)
            cfs_list[counter][0] = cf
            cfs_list[counter][1] = pose_index
            cfs_list[counter][2] = point_index
            counter += 1
    return cfs_list


@njit
def get_cf_main_clash(binding_site_grid, ligand_orientations, cf_size_list, n_cf_evals, load_cf_list,
                      ligand_atoms_types, default_cf, cell_width, min_xyz, load_range_list, preload_grid_distance,
                      clash_list, clash_list_size, num_atoms):
    cfs_list = np.zeros((n_cf_evals, 3), dtype=np.float32)
    counter = 0
    for point_index, point in enumerate(binding_site_grid):
        for pose_index, lig_pose in enumerate(ligand_orientations):
            cf = get_cf_with_clash(lig_pose, point, load_range_list, preload_grid_distance, cf_size_list, load_cf_list,
                                   ligand_atoms_types, default_cf, cell_width, min_xyz, clash_list, clash_list_size,
                                   num_atoms)
            cfs_list[counter][0] = cf
            cfs_list[counter][1] = pose_index
            cfs_list[counter][2] = point_index
            counter += 1
    return cfs_list


def main(target_name, preprocessed_target_path, preprocessed_ligand_path, result_folder_path, ligand_type='ligand',
         ligand_slice=None, write_info=True, unique_run_id=None, **user_config):
    """
    result_folder_path is the path to a folder in which the result and ligand pose folders will be created.
    """
    time_start = timeit.default_timer()
    save_time = False
    default_cf = 100000000
    info_lines = []
    output_lines = []

    params_dict_default = {
        'USE_CLASH': True,
        'LIGAND_ROTATIONS_PER_AXIS': 9,
        'LIGAND_TEST_DOT_SEPARATION': 1.5,
        'CLASH_DOT_DISTANCE': 0.25,
        'CONFORMERS_PER_MOLECULE': 1,
        'POSES_SAVED_PER_MOLECULE': 0,
        'WRITE_LIGAND_TEST_DOTS': False,
        'VERBOSE': False,
        'SAVE_TOTAL_TIME': False
    }
    params_dict = params_dict_default.copy()
    params_dict.update(user_config)

    test_dot_separation = params_dict['LIGAND_TEST_DOT_SEPARATION']
    conf_num = params_dict['CONFORMERS_PER_MOLECULE']
    poses_saved_per_molecule = params_dict["POSES_SAVED_PER_MOLECULE"]
    use_clash = params_dict["USE_CLASH"]
    ligand_rotations_per_axis = params_dict["LIGAND_ROTATIONS_PER_AXIS"]
    clash_dot_distance = params_dict["CLASH_DOT_DISTANCE"]
    write_ligand_test_dots = params_dict["WRITE_LIGAND_TEST_DOTS"]
    ligand_pose_save_path = os.path.join(result_folder_path, 'ligand_poses')

    if not os.path.isdir(result_folder_path):
        os.makedirs(result_folder_path)
    if not ligand_slice:
        ligand_slice = [0, None]
    start = ligand_slice[0]
    end = ligand_slice[1]
    conformers_per_molecule = int(os.path.basename(preprocessed_ligand_path).split('_')[2])
    output_file_name = f"{target_name}"
    if conformers_per_molecule > 1:
        output_file_name += f"_{conformers_per_molecule}_conf"
    if end:
        output_file_name += f"_split_{start}_{end}"
    if unique_run_id:
        output_file_name += f"_run_{unique_run_id}"
    output_file_path = os.path.join(result_folder_path, f'{output_file_name}.csv')
    counter = 2
    while os.path.isfile(output_file_path):
        output_file_path = os.path.join(result_folder_path, f'{output_file_name}_({counter}).csv')
        counter += 1

    binding_site_grid = np.load(os.path.join(preprocessed_target_path, f"ligand_test_dots_{test_dot_separation}.npy"))
    precalculated_cf_list = np.load(os.path.join(preprocessed_target_path, f"cf_list.npy"))

    if use_clash:
        load_range_list = np.load(os.path.join(preprocessed_target_path, "bd_site_cuboid_coord_range_array.npy"))
        clash_list = np.load(os.path.join(preprocessed_target_path, f"clash_list_{clash_dot_distance}.npy"))
        clash_list_size = clash_list.shape
    else:
        load_range_list = None
        clash_list = None
        clash_list_size = None

    if write_ligand_test_dots:
        write_pdb(binding_site_grid, "ligand_test_dots", ligand_pose_save_path, None, None)
    n_cf_evals = len(binding_site_grid) * ligand_rotations_per_axis**3
    atom_name_array, atom_type_array, atom_xyz_array, molecule_name_array, atoms_per_molecule_array, molecule_count_array \
        = load_ligands(preprocessed_target_path, ligand_type, start, end, conf_num, path_to_ligands=preprocessed_ligand_path)
    cfs_list_by_ligand = np.zeros(molecule_count_array, dtype=np.float32)
    if save_time:
        time_list = np.zeros(molecule_count_array, dtype=np.float32)

    cell_width = np.load(os.path.join(preprocessed_target_path, 'index_cube_cell_width.npy'))
    cf_size_list = np.array([np.size(precalculated_cf_list, axis=0), np.size(precalculated_cf_list, axis=1), np.size(precalculated_cf_list, axis=2)])

    info_lines.append(f"REMARK target folder: {preprocessed_target_path}")
    info_lines.append(f"REMARK software: {os.path.basename(__file__)}")
    info_lines.append(f"REMARK ligand type: {ligand_type}")
    info_lines.append(f"REMARK number of conformers: {conf_num}")
    info_lines.append(f"REMARK rotations per axis: {ligand_rotations_per_axis}")
    info_lines.append(f"REMARK dot separation: {test_dot_separation} A")
    info_lines.append(f"REMARK preloaded grid distance: {clash_dot_distance} A")
    info_lines.append(f"REMARK Total binding site grid dots: {len(binding_site_grid)}")
    info_lines.append(f"REMARK Total CF evaluations per ligand: {n_cf_evals}")
    info_lines.append(f"REMARK use clash: {use_clash}")
    info_lines.append(f"REMARK index cube width: {cell_width}")

    if params_dict['VERBOSE']:
        print("\n".join(info_lines))
    min_xyz = np.load(os.path.join(preprocessed_target_path, 'index_cube_min_xyz.npy'))

    for i, molecule in enumerate(atoms_per_molecule_array):
        time_molecule_start = timeit.default_timer()
        molecule_atom_count = atoms_per_molecule_array[i]
        molecule_atom_xyz = atom_xyz_array[i][0:molecule_atom_count]
        molecule_atom_types = atom_type_array[i][0:molecule_atom_count]
        molecule_rotations = rotate_ligand(molecule_atom_xyz, ligand_rotations_per_axis)
        num_atoms = len(molecule_rotations[0])
        if not use_clash:
            cfs_list = get_cf_main(binding_site_grid, molecule_rotations, cf_size_list, n_cf_evals, precalculated_cf_list,
                                   molecule_atom_types, default_cf, cell_width, min_xyz)
        else:
            cfs_list = get_cf_main_clash(binding_site_grid, molecule_rotations, cf_size_list, n_cf_evals,
                                         precalculated_cf_list, molecule_atom_types, default_cf, cell_width, min_xyz,
                                         load_range_list, clash_dot_distance, clash_list,
                                         clash_list_size, num_atoms)
        cfs_list_by_ligand[i] = np.min(cfs_list[:, 0])
        if poses_saved_per_molecule > 0:
            sorted_indices = np.argsort(cfs_list[:, 0])[:poses_saved_per_molecule]
            molecule_atoms_names = atom_name_array[i][0:molecule_atom_count]
            if poses_saved_per_molecule == 1:
                molecule_save_folder = ligand_pose_save_path
            else:
                molecule_save_folder = os.path.join(ligand_pose_save_path, molecule_name_array[i])
                if not os.path.isdir(molecule_save_folder):
                    os.makedirs(molecule_save_folder)
            for pose_number in range(0, poses_saved_per_molecule, 1):
                pose_index = sorted_indices[pose_number]
                pose_info = cfs_list[pose_index]
                pose_rotation_number = int(pose_info[1])
                translated_coords = np.zeros((len(molecule_rotations[pose_rotation_number]), 3), dtype=np.float32)
                for atom in range(len(molecule_rotations[pose_rotation_number])):
                    translated_coords[atom] = np.add(molecule_rotations[pose_rotation_number][atom],
                                                     binding_site_grid[int(pose_info[2])])
                pose_file_name = molecule_name_array[i]
                if conf_num == 1:
                    if pose_file_name.endswith('0'):
                        pose_file_name = pose_file_name.rsplit('_', 1)[0]
                if poses_saved_per_molecule != 1:
                    pose_file_name += f'_pose_{pose_number+1}'
                write_pdb(translated_coords, pose_file_name, molecule_save_folder, molecule_atoms_names,
                          [f"REMARK CF {cfs_list[sorted_indices[pose_number]][0]:.2f}\n",
                           f"REMARK atom types: {np.array2string(molecule_atom_types, separator=' ', max_line_width=2000).strip('[]')}\n"])
        if save_time:
            time_list[i] = timeit.default_timer() - time_molecule_start

    if params_dict['SAVE_TOTAL_TIME']:
        info_lines.append(f"REMARK total screen time: {timeit.default_timer() - time_start:.3f} seconds")
        if params_dict['VERBOSE']:
            print(output_lines[-1])

    if write_info:
        info_file_path = os.path.splitext(output_file_path)[0] + f'_info.txt'
        with open(info_file_path, "w") as f:
            f.writelines("\n".join(info_lines))

    output_header = "Name,CF"
    if ligand_type != 'ligand':
        output_header += ',Type'
    if conf_num > 1:
        output_header += ",Conformer_number"
    if save_time:
        output_header += ",Time"
    else:
        output_lines.append(output_header)
    for z, ligand in enumerate(atoms_per_molecule_array):
        output = f"{molecule_name_array[z].rsplit('_', 1)[0]},{cfs_list_by_ligand[z]:.0f}"
        if ligand_type != 'ligand':
            output += f',{ligand_type}'
        if conf_num > 1:
            try:
                output += "," + molecule_name_array[z].rsplit('_', 1)[1]
            except:
                output += f",{str(0)}"
        if save_time:
            output += f",{time_list[z]:.3f}"
        output_lines.append(output)
    with open(output_file_path, "w") as f:
        f.writelines("\n".join(output_lines))
        f.write("\n")
    if params_dict['VERBOSE']:
        print("\n".join(output_lines))
    return output_file_path


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", '--path_to_target', required=True, type=str, help='Path to target folder')
    parser.add_argument("-s", '--ligand_slice', type=str, help='Subset of ligands to screen. 2 integers separated by a comma (last one not considered): 0,100')
    parser.add_argument('-l', '--ligand_path', default=None, type=str, help='Custom ligand path')
    parser.add_argument('-lt', '--ligand_type', type=str, default='ligand',help='Specify if ligand is of a special type')
    parser.add_argument('-si', '--skip_info', action='store_true', help='Skips writing a target info file')
    parser.add_argument('-o', '--create_folder', action='store_false', help='Prevents nrgrank from making folders to store output')
    parser.add_argument('-t', '--temp_path', default=None, type=str, help='Custom temp path')

    args = parser.parse_args()
    path_to_target = args.path_to_target
    category = args.ligand_type
    ligand_slice = args.ligand_slice
    if ligand_slice:
        ligand_slice = [int(x) for x in ligand_slice.split(',')]
    path_to_ligands = args.ligand_path
    skip_info = args.skip_info
    skip_remark = args.skip_remark
    create_folder = args.create_folder
    temp_path = args.temp_path

    main(path_to_target, path_to_ligands, category, ligand_slice, skip_info, skip_remark, create_folder,
         temp_path=temp_path)


if __name__ == "__main__":
    get_args()



from .process_target import main as process_target
from .process_ligands import main as process_ligands
from .rank_molecules import main as nrgrank_main
from .generate_conformers import main as generate_conformers


import os
import json
import importlib.resources


def write_pdb(coord_list, name, path, ligand_names, extra_info):
    if not os.path.isdir(os.path.dirname(path)):
        os.mkdir(os.path.dirname(path))
    if not os.path.isdir(path):
        os.mkdir(path)
    textfile = open(os.path.join(path, name + ".pdb"), 'w')
    counter = 0
    if extra_info is not None:
        for line in extra_info:
            textfile.write(line)
    for line in coord_list:
        if ligand_names is not None:
            textfile.write("HETATM {:>4}  {:<4}LIG L   1{:>12} {:>7} {:>7}  1.00  0.10 \n".format(str(counter),
                                                                                                    ligand_names[counter],
                                                                                                    str(round(line[0], 3)),
                                                                                                    str(round(line[1], 3)),
                                                                                                    str(round(line[2], 3))))
        else:
            textfile.write("HETATM {:>4}  C   DOT X   1{:>12} {:>7} {:>7}  1.00  0.10 \n".format(str(counter),
                                                                                                  str(round(line[0], 3)),
                                                                                                  str(round(line[1], 3)),
                                                                                                  str(round(line[2], 3))))
        counter += 1
    textfile.close()


def load_rad_dict():
    file_path = importlib.resources.files('nrgrank').joinpath('deps', 'atom_type_radius.json')
    with file_path.open('r') as file:
        loaded_atom_data = json.load(file)
    return loaded_atom_data


def get_radius_number(letter_type, rad_dict):
    letter_type = letter_type.upper().replace(' ', '')
    try:
        atm_info = [rad_dict[letter_type]['type_number'], rad_dict[letter_type]['radius']]
    except KeyError:
        atm_info = [39, 2.00]
    atm_type_num = atm_info[0]
    atm_rad = atm_info[1]
    return atm_type_num, atm_rad


import os
import numpy as np
from numba import njit
from nrgrank.general_functions import load_rad_dict, get_radius_number
import shutil
import timeit
import argparse
from datetime import date
import importlib.resources


def load_atoms_mol2(mol2_file_path, rad_dict):
    coord_start = 0
    atoms_xyz = []
    atoms_numbers = []
    atoms_types = []
    atoms_radius = []
    with open(mol2_file_path) as f:
        for line in f:
            if line.strip():
                line = line.split()
                if line[0] == '@<TRIPOS>ATOM':
                    coord_start = 1
                if line[0][0] == '@' and coord_start == 1 and line[0] != '@<TRIPOS>ATOM':
                    break
                if coord_start == 1 and line[0] != '@<TRIPOS>ATOM' and line[5].split(".")[0] != 'H':
                    atoms_xyz.append([float(line[2]), float(line[3]), float(line[4])])
                    atoms_numbers.append(int(line[0]))
                    atoms_type, atom_radius = get_radius_number(line[5], rad_dict)
                    atoms_types.append(atoms_type)
                    atoms_radius.append(atom_radius)
    atoms_xyz = np.array(atoms_xyz, dtype=np.float32)
    atoms_radius = np.array(atoms_radius, dtype=np.float32)
    atoms_numbers = np.array(atoms_numbers, dtype=np.int32)
    atoms_types = np.array(atoms_types, dtype=np.int32)
    atoms_types_sorted = atoms_types[atoms_numbers.argsort()]

    return atoms_xyz, atoms_types_sorted, atoms_radius


def load_binding_site_pdb(binding_site):
    atom_coord_type_list = []
    with open(binding_site) as f:
        lines = f.readlines()
        for line in lines:
            if line.startswith('ATOM'):
                line = line.strip()
                temp_array = [float(line[30:38]), float(line[38:46]), float(line[46:54]), float(line[60:66])]  # sphere radius
                atom_coord_type_list.append(temp_array)
    return atom_coord_type_list


def make_binding_site_cuboid(dot_division, a, padding, preprocessed_file_path):
    # find min and max coords for box as well as remove or add sphere radius
    x = [np.min(a[:, 0]) - a[np.argmin(a[:, 0]), 3] - padding, np.max(a[:, 0]) + a[np.argmax(a[:, 0]), 3] + padding]
    y = [np.min(a[:, 1]) - a[np.argmin(a[:, 1]), 3] - padding, np.max(a[:, 1]) + a[np.argmax(a[:, 1]), 3] + padding]
    z = [np.min(a[:, 2]) - a[np.argmin(a[:, 2]), 3] - padding, np.max(a[:, 2]) + a[np.argmax(a[:, 2]), 3] + padding]
    bd_site_box = []
    for x_coord in x:
        for y_coord in y:
            for z_coord in z:
                bd_site_box.append([x_coord, y_coord, z_coord])
    # write_pdb(bd_site_box, f"bd_site_box", './temp/ligand_poses', None, None)
    x_range = np.arange(x[0], x[1], dot_division)
    y_range = np.arange(y[0], y[1], dot_division)
    z_range = np.arange(z[0], z[1], dot_division)
    np.save(os.path.join(preprocessed_file_path, "bd_site_cuboid_coord_range_array"), np.stack((x, y, z)))
    return x_range, y_range, z_range


def build_index_cubes(water_vdw_radius, target_atoms_xyz, atoms_radius, preprocessed_file_path, cw_factor=1, custom_cell_width=False):
    grid_placeholder = -1
    max_rad = np.amax(atoms_radius, axis=0)
    cell_width = 2 * (max_rad + water_vdw_radius)
    if custom_cell_width:
        cell_width = custom_cell_width
    max_xyz = np.zeros(3)
    max_xyz[0] = np.max(target_atoms_xyz[:, 0]) + cell_width*cw_factor
    max_xyz[1] = np.max(target_atoms_xyz[:, 1]) + cell_width*cw_factor
    max_xyz[2] = np.max(target_atoms_xyz[:, 2]) + cell_width*cw_factor
    min_xyz = np.zeros(3)
    min_xyz[0] = np.min(target_atoms_xyz[:, 0]) - cell_width*cw_factor
    min_xyz[1] = np.min(target_atoms_xyz[:, 1]) - cell_width*cw_factor
    min_xyz[2] = np.min(target_atoms_xyz[:, 2]) - cell_width*cw_factor
    lengths = ((max_xyz - min_xyz) / cell_width).astype(np.int32) + 1
    temp_grid = []
    for i in range(lengths[0]):
        temp_grid.append([])
        for j in range(lengths[1]):
            temp_grid[i].append([])
            for k in range(lengths[2]):
                temp_grid[i][j].append([])
    for i, row in enumerate(target_atoms_xyz):
        grid_indices = ((row[:3] - min_xyz) / cell_width).astype(np.int32)
        temp_grid[grid_indices[0]][grid_indices[1]][grid_indices[2]].append(i)
    max_cell_len = 0
    for row in temp_grid:
        for col in row:
            for cell in col:
                n = len(cell)
                if n > max_cell_len:
                    max_cell_len = n
    grid = np.full((lengths[0], lengths[1], lengths[2], max_cell_len), grid_placeholder, dtype=np.int32)
    for i in range(lengths[0]):
        for j in range(lengths[1]):
            for k in range(lengths[2]):
                for x, v in enumerate(temp_grid[i][j][k]):
                    grid[i][j][k][x] = v
    np.save(os.path.join(preprocessed_file_path, f"index_cube_min_xyz"), min_xyz)
    np.save(os.path.join(preprocessed_file_path, f"index_cube_cell_width"), cell_width)
    coord_list = []
    name_list = []
    for x in np.arange(min_xyz[0], max_xyz[0], cell_width):
        for y in np.arange(min_xyz[1], max_xyz[1], cell_width):
            for z in np.arange(min_xyz[2], max_xyz[2], cell_width):
                coord_list.append([x, y, z])
                if not name_list:
                    name_list = ["O"]
                elif name_list == ["O"]:
                    name_list.append("N")
                else:
                    name_list.append("C")
    # write_pdb(coord_list, "binding_grid_test", './temp/ligand_poses', name_list, None)
    return grid, min_xyz, cell_width, max_xyz


def prepare_preprocess_output(path_to_target, params_dict):
    numpy_output_path = os.path.join(path_to_target, 'preprocessed_target')
    os.makedirs(numpy_output_path, exist_ok=True)
    config_output = os.path.join(numpy_output_path, "config.txt")
    with open(config_output, "w") as config_file:
        config_file.write(f"DATE_PREPARED {date.today().strftime('%d/%m/%Y')}\n")
        for parameter in params_dict:
            config_file.write(f"{parameter}={params_dict[parameter]}\n")
    return numpy_output_path


def load_ligand_test_dots(test_dot_separation, binding_site_spheres):
    """ This function uses the binding site spheres to make dots on which the ligand will be centered for testing poses"""
    grid_coords = []
    a = np.array(binding_site_spheres)

    x = [round(np.min(a[:, 0]) - a[np.argmin(a[:, 0]), 3], 3), round(np.max(a[:, 0]) + a[np.argmax(a[:, 0]), 3], 3)]
    y = [round(np.min(a[:, 1]) - a[np.argmin(a[:, 1]), 3], 3), round(np.max(a[:, 1]) + a[np.argmax(a[:, 1]), 3], 3)]
    z = [round(np.min(a[:, 2]) - a[np.argmin(a[:, 2]), 3], 3), round(np.max(a[:, 2]) + a[np.argmax(a[:, 2]), 3], 3)]

    for dot_x in np.arange(x[0], x[1], test_dot_separation):
        for dot_y in np.arange(y[0], y[1], test_dot_separation):
            for dot_z in np.arange(z[0], z[1], test_dot_separation):
                coords = np.array([round(dot_x, 3), round(dot_y, 3), round(dot_z, 3)])
                for row in a:
                    distance = np.linalg.norm(coords - row[:3])
                    if distance < row[3]:
                        grid_coords.append(coords)
                        break
    return grid_coords


def clean_binding_site_grid(target_grid, binding_site_grid, min_xyz, cell_width, target_atoms_xyz, ligand_test_dot_file_path):
    index = []
    for a, point in enumerate(binding_site_grid):
        grid_index = ((point - min_xyz) / cell_width).astype(np.int32)
        for i_offset in [-1, 0, 1]:
            for j_offset in [-1, 0, 1]:
                for k_offset in [-1, 0, 1]:
                    i = i_offset + grid_index[0]
                    j = j_offset + grid_index[1]
                    k = k_offset + grid_index[2]
                    if i < len(target_grid) and j < len(target_grid[0]) and k < len(target_grid[0][0]):
                        for neighbour in target_grid[i][j][k]:
                            if neighbour == -1:
                                break
                            else:
                                dist = np.sqrt((target_atoms_xyz[neighbour][0] - point[0]) ** 2 +
                                               (target_atoms_xyz[neighbour][1] - point[1]) ** 2 +
                                               (target_atoms_xyz[neighbour][2] - point[2]) ** 2)
                                if dist <= 2.0:
                                    index.append(a)

                                    break
    cleaned_binding_site_grid = np.delete(binding_site_grid, index, 0)
    # write_pdb(cleaned_binding_site_grid, "cleaned_grid", f'./temp/ligand_poses/', None, None)
    np.save(ligand_test_dot_file_path, cleaned_binding_site_grid)


@njit
def get_cf_list(target_grid, atom_type_range, target_atom_types, energy_matrix, number_types):
    target_grid_x = len(target_grid)
    target_grid_y = len(target_grid[0])
    target_grid_z = len(target_grid[0][0])
    result_array = np.zeros((target_grid_x, target_grid_y, target_grid_z, number_types))
    for x in np.arange(0, target_grid_x):
        for y in np.arange(0, target_grid_y):
            for z in np.arange(0, target_grid_z):
                grid_index = [x, y, z]
                for counter, atom_type in enumerate(atom_type_range):
                    cf = 0.0
                    for i_offset in [-1, 0, 1]:
                        for j_offset in [-1, 0, 1]:
                            for k_offset in [-1, 0, 1]:
                                i = i_offset + grid_index[0]
                                j = j_offset + grid_index[1]
                                k = k_offset + grid_index[2]
                                if 0 < i < len(target_grid) and 0 < j < len(target_grid[0]) and 0 < k < len(target_grid[0][0]):
                                    if target_grid[i][j][k][0] != -1:
                                        for neighbour in target_grid[i][j][k]:
                                            if neighbour == -1:
                                                break
                                            else:
                                                # Normal program:
                                                type_1 = atom_type
                                                type_2 = target_atom_types[neighbour]
                                                # Randomly chosen atom type
                                                # type_1 = np.random.randint(1, 41)
                                                # type_2 = np.random.randint(1, 41)
                                                energy_value = energy_matrix[type_1][type_2]
                                                cf += energy_value
                    result_array[x, y, z, counter] = cf
    return result_array


@njit
def get_clash_per_dot(x_range, y_range, z_range, target_grid, min_xyz, cell_width, target_atoms_xyz, max_size_array):
    clash_list = np.zeros((max_size_array[0], max_size_array[1], max_size_array[2]), dtype=np.bool_)
    for a, x_value in enumerate(x_range):
        for b, y_value in enumerate(y_range):
            for c, z_value in enumerate(z_range):
                ligand_atom = np.array([x_value, y_value, z_value], dtype=np.float32)
                clash_list[a][b][c] = get_clash_for_dot(ligand_atom, target_grid, min_xyz, cell_width, target_atoms_xyz)
    return clash_list


@njit
def get_clash_for_dot(ligand_atom_coord, target_grid, min_xyz, cell_width, target_atoms_xyz):
    # TODO: test if clashes per radius associated to each type is better
    clash = False
    grid_index = ((ligand_atom_coord - min_xyz) / cell_width).astype(np.int32)
    for i_offset in [-1, 0, 1]:
        for j_offset in [-1, 0, 1]:
            for k_offset in [-1, 0, 1]:
                i = i_offset + grid_index[0]
                j = j_offset + grid_index[1]
                k = k_offset + grid_index[2]
                if 0 <= i < len(target_grid) and 0 <= j < len(target_grid[0]) and 0 <= k < len(target_grid[0][0]):
                    if target_grid[i][j][k][0] != -1:
                        for neighbour in target_grid[i][j][k]:
                            if neighbour == -1:
                                break
                            else:
                                dist = np.linalg.norm(target_atoms_xyz[neighbour] - ligand_atom_coord)
                                if dist <= 2.0:
                                    clash = True
                                    return clash
    return clash


def preprocess_one_target(target_file_path, binding_site_file_path, params_dict, energy_matrix, time_start, overwrite,
                          verbose=False, create_new_dir=True):
    use_clash = params_dict["USE_CLASH"]
    clash_dot_distance = params_dict['CLASH_DOT_DISTANCE']
    bd_site_cuboid_padding = params_dict["BD_SITE_CUBOID_PADDING"]
    cell_width = params_dict['CELL_WIDTH']
    test_dot_separation = params_dict['LIGAND_TEST_DOT_SEPARATION']
    water_vdw_radius = params_dict['WATER_RADIUS']

    target = os.path.splitext(os.path.basename(target_file_path))[0]
    if create_new_dir:
        target_save_dir = os.path.join(os.path.dirname(target_file_path), f"{target}_NRGRank")
        os.makedirs(target_save_dir, exist_ok=True)
        new_target_file_path = os.path.join(target_save_dir, 'target.mol2')
        shutil.copyfile(target_file_path, new_target_file_path)
        target_file_path = new_target_file_path
        new_bd_site_path = os.path.join(target_save_dir, 'bd_site.pdb')
        shutil.copyfile(binding_site_file_path, new_bd_site_path)
        binding_site_file_path = new_bd_site_path
    else:
        target_save_dir = os.path.dirname(target_file_path)
    if not os.path.isfile(target_file_path):
        exit(f"Could not find target file at path: {target_file_path}")
    if not os.path.isfile(binding_site_file_path):
        exit(f"Could not find binding site file at path: {binding_site_file_path}")

    # ####################### DEFINE CUBOID AROUND BINDING SITE #######################

    rad_dict = load_rad_dict()
    number_of_atom_types = len(energy_matrix)-2
    target_atoms_xyz, target_atoms_types, atoms_radius = load_atoms_mol2(target_file_path, rad_dict)
    preprocessed_target_folder_path = prepare_preprocess_output(target_save_dir, params_dict)
    index_cubes, min_xyz, cell_width, max_xyz = build_index_cubes(water_vdw_radius, target_atoms_xyz, atoms_radius,
                                                                  preprocessed_target_folder_path,
                                                                  custom_cell_width=cell_width)
    binding_site_spheres = load_binding_site_pdb(binding_site_file_path)
    binding_site_x_range, binding_site_y_range, binding_site_z_range = make_binding_site_cuboid(clash_dot_distance,
                                                                                                np.array(binding_site_spheres),
                                                                                                bd_site_cuboid_padding,
                                                                                                preprocessed_target_folder_path)

    # ####################### PRECALCULATE CF #######################

    cf_array_path = os.path.join(preprocessed_target_folder_path, "cf_list.npy")
    clash_file_path = os.path.join(preprocessed_target_folder_path, f"clash_list_{clash_dot_distance}.npy")
    if not os.path.isfile(cf_array_path) or overwrite:
        max_size_array = np.array([len(binding_site_x_range), len(binding_site_y_range), len(binding_site_z_range), number_of_atom_types], dtype=np.int32)
        atom_type_range = np.arange(1, number_of_atom_types+1)
        if use_clash:
            if verbose:
                print('Getting clashes')
            clash_list = get_clash_per_dot(binding_site_x_range, binding_site_y_range, binding_site_z_range,
                                           index_cubes, min_xyz, cell_width, target_atoms_xyz, max_size_array)
            np.save(clash_file_path, clash_list)
        if verbose:
            print('Precalculating CF')
        cfs_list = get_cf_list(index_cubes, atom_type_range, target_atoms_types, energy_matrix, number_of_atom_types)
        np.save(cf_array_path, cfs_list)
    else:
        print(f"Energies already precalculated... Skipping. \nUse -o flag if you wish to overwrite.")

    # ####################### GENERATE AND CLEAN LIGAND TEST DOTS #######################

    ligand_test_dot_file_path = os.path.join(preprocessed_target_folder_path, f"ligand_test_dots_{test_dot_separation}.npy")
    if not os.path.isfile(ligand_test_dot_file_path) or overwrite:
        original_grid = load_ligand_test_dots(test_dot_separation, binding_site_spheres)
        clean_binding_site_grid(index_cubes, original_grid, min_xyz, cell_width, target_atoms_xyz,
                                ligand_test_dot_file_path)
    else:
        if verbose:
            print(f"The file for binding site dots at {test_dot_separation} A distance already exists")
    if verbose:
        total_run_time = timeit.default_timer() - time_start
        if total_run_time > 60.0:
            total_run_time /= 60
            print(f"{target}: {total_run_time} minutes to run")
        else:
            print(f"{target}: {total_run_time:.2f} seconds to run")
    return preprocessed_target_folder_path


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", '--path_to_targets', required=True, type=str,
                        help='Path to folder containing target folders')
    parser.add_argument('-t', '--specific_target', type=str,
                        help='Specify target(s) to analyse. If multiple: separate with comma no space')
    parser.add_argument('-o', '--overwrite', action='store_true',
                        help='specify if you want to overwrite pre existing files')

    args = parser.parse_args()
    path_to_targets = os.path.abspath(args.path_to_targets)
    if args.specific_target is None:
        target_list = next(os.walk(path_to_targets))[1]
    else:
        target_list = args.specific_target.split(',')
    target_list = sorted(target_list)
    main(path_to_targets, target_list, overwrite=args.overwrite)


def main(target_mol2_path, binding_site_file_path, create_new_dir=True, overwrite=False, **user_config):
    time_start = timeit.default_timer()
    params_dict_default = {
        'WATER_RADIUS': 1.4,
        'MATRIX_NAME': 'MC_5p_norm_P10_M2_2_multiplied_2',
        'CLASH_DOT_DISTANCE': 0.25,
        'BD_SITE_CUBOID_PADDING': 2,
        'LIGAND_TEST_DOT_SEPARATION': 1.5,
        'USE_CLASH': True,
        'CELL_WIDTH': 6.56,
        'VERBOSE': False
    }
    params_dict = params_dict_default.copy()
    params_dict.update(user_config)
    use_clash = params_dict['USE_CLASH']
    matrix_name = params_dict['MATRIX_NAME']
    verbose = params_dict['VERBOSE']
    if not use_clash:
        print('Considering poses with clashes')

    matrix_path = importlib.resources.files('nrgrank').joinpath('deps', 'matrix', f'{matrix_name}.npy')
    energy_matrix = np.load(matrix_path)

    if not type(target_mol2_path) is str or type(binding_site_file_path) is not str:
        exit('target_file and binding_site_file_path must be strings')

    if not os.path.isfile(target_mol2_path):
        exit(f'{target_mol2_path} is not a file')
    if not os.path.isfile(binding_site_file_path):
        exit(f'{binding_site_file_path} is not a file')

    target_save_dir = preprocess_one_target(target_mol2_path, binding_site_file_path, params_dict, energy_matrix,
                                            time_start, overwrite, verbose, create_new_dir)
    return target_save_dir


if __name__ == "__main__":
    get_args()
